(* +------------------------------------------------------------------------+
   |                                                                        |
   |                                                                        |
   |                               Manipulation d'automates                 |
   |                                                                        |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

(* Michel Quercia, le 10/01/98 *)

#open "graphx";;
#open "graphes";;
#open "arbres";;
#open "map";;


             (* +-------------------------------------+
                |  Utilitaires sur les listes tries  |
                +-------------------------------------+ *)

(* union de liste tries *)
let rec prefix ++ e f = match (e,f) with
| ([],_) -> f
| (_,[]) -> e
| (x::x',y::y') -> if x < y      then x :: x' ++ f
                   else if x > y then y :: e ++ y'
                   else x :: x' ++ y'
;;

(* retire un lment d'une liste trie *)
let rec lrem l x = match l with
| [] -> []
| y :: suite when x = y -> suite
| y :: suite -> y :: (lrem suite x)
;;



                 (* +----------------------------+
                    |  Construction d'automates  |
                    +----------------------------+ *)


let new_automate() = {g = new_graphe(); initial = []; final = []};;
let mirroir(a) = {g = graphes__mirroir a.g; initial = a.final; final = a.initial};;

let add_tat a x e =
  { g = add_noeud a.g x e; initial = a.initial; final = a.final }
;;

let del_tat a x =
  { g = del_noeud a.g x; initial = a.initial; final = a.final }
;;

(* ajoute une transition x -- c --> y *)
let add_trans a (x,y) c =
  let s = match arc_opt a.g (x,y) with
  | None    -> [c]
  | Some(s) -> [c] ++ s
  in
  { g = add_arc a.g (x,y) s; initial = a.initial; final = a.final }
;;

(* supprime une transition x -- c --> y *)
let del_trans a (x,y) c =
  let s = match arc_opt a.g (x,y) with
  | None    -> []
  | Some(s) -> lrem s c
  in
  { g = if s = [] then del_arc a.g (x,y) else add_arc a.g (x,y) s;
    initial = a.initial;
    final = a.final }
;;


                    (* +-----------------------+
                       |  Renommage des tats  |
                       +-----------------------+ *)

let map_auto f a =
  let m = ref(empty eq__compare) and num = ref(-1) in
  do_noeuds (fun (x,e) -> m := add x (f(x,e)) !m) a.g;
  let ff(x) = fst(find x !m) in

  let g = ref(new_graphe()) in
  iter (fun _ (x,e) -> g := add_noeud !g x e) !m;
  do_arcs (fun ((x,y),c) -> g := add_arc !g (ff x,ff y) c) a.g;

  {g = !g; initial = map ff a.initial; final = map ff a.final}
;;


let map_int a =
  let num = ref(-1) in
  let f _ = incr num; (!num,string_of_int !num) in
  map_auto f a
;;



                   (* +-------------------------+
                      |  Cltures, transitions  |
                      +-------------------------+ *)

let cltures(a) =

  (* initialise la clture de x  [x] *)
  let cl = ref(empty eq__compare) in
  do_noeuds (fun (x,_) -> cl := add x [x] !cl) a.g;

  (* clture inverse *)
  let rv = ref(!cl) in

  (* traitement d'une transition spontane : si y n'est pas dj dans *)
  (* dans la clture de x, fusionne la clture de y avec celles des   *)
  (* antcdents de x.                                                *)
  let job = fun
  | ((x,y),Epsilon::_) -> if not (list__mem y (find x !cl)) then begin
     let cy = find y !cl and rx = find x !rv in
     do_list (fun z -> cl := add z ((find z !cl) ++ cy) (remove z !cl)) rx;
     do_list (fun z -> rv := add z ((find z !rv) ++ rx) (remove z !rv)) cy
  end
  | _ -> ()
  in do_arcs job a.g;

  (* rsultat *)
  !cl
;;

let avance a cl tats c =
  let l = ref(if c = Epsilon then sort__sort (prefix <=) tats else []) in
  let job x y = if list__mem c (arc a.g (x,y)) then l := !l ++ (map__find y cl) in
  do_list (fun x -> do_list (job x) (succ a.g x)) tats;
  !l
;;

                      (* +-------------------+
                         |  Dterminisation  |
                         +-------------------+ *)


(* cherche l'ensemble des successeurs d'un ensemble d'tats *)
(* g = graphe de transition, cl = cltures, ens = ensemble initial complet *)
let succ_ens g cl ens =

  let m = ref(empty eq__compare) in

  (* ajoute la clture de y pour toutes les transitions non vides x -> y *)
  let job x y =
    do_list (fun Epsilon -> ()
               | c -> let s = (try find c !m with Not_found -> []) in
                      if not(mem y s) then m := add c (s ++ (find y cl)) !m)
            (arc g (x,y))
  in do_list (fun x -> do_list (job x) (succ g x)) ens;

  (* rsultat = table lettre -> ensemble *)
  !m
;;

(*
 Algorithme de dterminisation :
 a = automate,
 f = tiquetage des listes codes
*)
let dterm f a =

  let cl = cltures(a)
  and b = ref{g = new_graphe(); initial = []; final = []} in

  (* ajoute un  un les ensembles d'tats, retourne l'ensemble cod *)
  let rec job(e) =

    if not(member !b.g e) then begin
      b := add_tat !b e (f e);
      if exists (fun x -> list__mem x a.final) e
      then b := { g = !b.g; initial = []; final = e :: !b.final };
      iter (fun c e1 -> job(e1); b := add_trans !b (e,e1) c) (succ_ens a.g cl e)
    end

  in 
  let i = avance a cl a.initial Epsilon in
  job(i);
  { g = !b.g; initial = [i]; final = !b.final}
;;

let dterminise(a) = dterm (fun e -> concat (map (etiq a.g) e)) a;;


(* algorithme de minimisation : mirroir -> dterminise -> mirroir -> dterminise *)
let minimise(a) =
  let dterm(a) = dterm (fun _ -> ()) a in
  map_int( dterm (mirroir (dterm (mirroir a))))
;;

                  (* +--------------------------+
                     |  Expressions rgulires  |
                     +--------------------------+ *)

(* dsimbrique des unions ou des concatnations *)
let merge_union = fun
| (Union x) (Union y) 	-> Union (x@y)
| (Union x) y         	-> Union(x@[y])
| x (Union y)         	-> Union(x::y)
| x y                 	-> Union [x;y]
and merge_concat = fun
| (Concat x) (Concat y) -> Concat (x@y)
| (Concat x) y          -> Concat(x@[y])
| x (Concat y)          -> Concat(x::y)
| x y                   -> Concat [x;y]
;;

(* analyseur d'entre *)
#open "genlex";;
let expression = 
  let rec lexeur(s) = match s with
      | [< '` `|`\t`|`\n` >] -> lexeur(s)
      | [< '`|` >] -> [< 'Infixe(1,merge_union);  lexeur s >]
      | [< '`.` >] -> [< 'Infixe(2,merge_concat); lexeur s >]
      | [< '`*` >] -> [< 'Postfixe(Etoile);       lexeur s >]
      | [< '`+` >] -> [< 'Postfixe(Plus);         lexeur s >]
      | [< '`@` >] -> [< 'Valeur(Const(Epsilon)); lexeur s >]
      | [< '`(` >] -> [< 'Lpar; 	   	  lexeur s >]
      | [< '`)` >] -> [< 'Rpar; 	   	  lexeur s >]
      | [< '`<`; '`>` >]     -> [< 'Valeur(Vide); lexeur s >]
      | [< '`\\`; echap c >] -> [< 'Valeur(Const(Lettre c)); lexeur s >]
      | [< 'c >]             -> [< 'Valeur(Const(Lettre c)); lexeur s >]
      | [< >] -> [< >]
  and echap = function
      | [< chiffre a; chiffre b; chiffre c >] -> char_of_int(100*a+10*b+c)
      | [< '`n` >] -> `\n`
      | [< '`t` >] -> `\t`
      | [< '`b` >] -> `\b`
      | [< '`r` >] -> `\r`
      | [< 'c >] -> c
  and chiffre = function
      | [< '`0`..`9` as c >] -> int_of_char(c) - 48
  in let eval = analyseur (fun () -> (2,merge_concat)) (fun x -> x) lexeur
  in function s -> try eval s
                   with Erreur(ex,i) -> print_erreur s i; raise ex
;;
#close "genlex";;

(* impression *)
let rec g_arbre_of_expression = function
| Vide             -> G_noeud("<>",[])
| Const(Epsilon)   -> G_noeud("@",[])
| Const(Lettre(l)) -> G_noeud(make_string 1 l,[])
| Union(l)         -> G_noeud("|", map g_arbre_of_expression l)
| Concat(l)        -> G_noeud(".", map g_arbre_of_expression l)
| Etoile(e)        -> G_noeud("*", [g_arbre_of_expression e])
| Plus(e)          -> G_noeud("+", [g_arbre_of_expression e])
;;

let print_expression e = print_string_g_arbre(g_arbre_of_expression e);;


           (* +-----------------------------------------+
              |  transformation expression -> automate  |
              +-----------------------------------------+ *)

(* adjonction d'tats et de transitions *)
let addn a n = 	   a := add_tat !a n (string_of_int n)
and addt a n p c = a := add_trans !a (n,p) c
;;

(* conditions d'entre :
 * a (i,n)
 * a = rfrence sur l'automate en cours
 * i = tat initial impos
 * n = 1er numro d'tat disponible
 *
 * (f,n)
 * f = tat de sortie produit
 * n = 1er numro d'tat disponible
 *
 * Attention : ne pas produire de transition vers un tat initial !
 *)

let rec auto_of_exp a (i,n) = function

| (Vide|Union [])            -> addn a n; (n,n+1)
| (Const(Epsilon)|Concat []) -> (i,n)
| Const(l)                   -> addn a n; addt a i n l; (n,n+1) 

| Union([u]) ->
  let (f1,n1) = auto_of_exp a (i,n) u in
  addn a n1; addt a f1 n1 Epsilon; (n1,n1+1)

| Union(u::suite) ->
  let (f1,n1) = auto_of_exp a (i,n)  u             in
  let (f2,n2) = auto_of_exp a (i,n1) (Union suite) in
  addt a f1 f2 Epsilon; (f2,n2)

| Concat(u::suite) ->
  let (f1,n1) = auto_of_exp a (i,n)   u              in
  let (f2,n2) = auto_of_exp a (f1,n1) (Concat suite) in
  (f2,n2)

| Etoile(u) ->
  addn a n; addt a i n Epsilon;
  let (f1,n1) = auto_of_exp a (n,n+1) u in
  addt a f1 n Epsilon; (n,n1)

| Plus(u) ->
  addn a n; addt a i n Epsilon;
  let (f1,n1) = auto_of_exp a (n,n+1) u in
  addt a f1 n Epsilon; (f1,n1)
;;

let automate_of_expression(e) =
  let a = ref{g = new_graphe();
              initial = [0]; 
              final   = []} in
  addn a 0;
  let (f,_) = auto_of_exp a (0,1) e in
  {g = !a.g; initial = [0]; final = [f]}
;;


         (* +--------------------------------------------+
            |  Simplification de l'automate de Thompson  |
            +--------------------------------------------+ *)


(* supprime tous les tats non finals n'ayant pas de transition sur lettre *)
(* sortante et tous les tats non initials n'ayant pas de transition sur   *)
(* lettre entrante. Supprime aussi les transitions vides d'un tat vers    *)
(* lui-mme.                                                               *)
let simplifie auto =

  let g = ref(auto.g)
  and i = ref(auto.initial)
  and f = ref(auto.final)    in

  (****************** limination des tats inutiles *********************)
  let job(e,_) = let s = succ !g e and p = pred !g e in

    (* cherche s'il existe une transition non vide partant de e *)
    let sortie() = try
      do_list (fun x -> if arc !g (e,x) <> [Epsilon] then raise Exit) s;
      list__mem e !f
    with Exit -> true in

    (* cherche s'il existe une transition non vide arrivant en e *)
    let entre() = try
      do_list (fun x -> if arc !g (x,e) <> [Epsilon] then raise Exit) p;
      list__mem e !i
    with Exit -> true in

    (* court-circuite tous les arcs passant par e *)
    let racourcit x y =
      let s = match arc !g (x,e) with
      | [Epsilon] -> arc !g (e,y)
      | s         -> s
      and s' = match arc_opt !g (x,y) with
      | None -> []
      | Some(t) -> t
      in g := add_arc !g (x,y) (s ++ s')
    in

    if not(entre() & sortie()) then begin

      do_list (fun x -> do_list (racourcit x) s) p;

      (* si e est initial, ajoute les successeurs de e dans !i *)
      if list__mem e !i then begin
    	do_list (fun x -> i := !i ++ [x]) s;
    	i := lrem !i e
      end;
  
      (* si e est final, ajoute les prdecesseurs de e dans !f *)
      if list__mem e !f then begin
    	do_list (fun x -> f := !f ++ [x]) p;
    	f := lrem !f e
      end;
  
      (* supprime l'tat e *)
      g := del_noeud !g e

    end in
  do_noeuds job auto.g;

  (**************** limination des boucles vides *******************)
  let job(e,_) = match arc_opt !g (e,e) with
  | Some([Epsilon])  -> g := del_arc !g (e,e)
  | Some(Epsilon::s) -> g := add_arc !g (e,e) s
  | _ -> ()
  in do_noeuds job !g;

  (* termin *)
  { g = !g; initial = !i; final = !f}
;;

                      (* +-------------------+
                         |  E/S d'automates  |
                         +-------------------+ *)



let output_automate channel a =
  let l = ref [] in
  do_noeuds (fun x -> l := x :: !l) a.g;
  output_value channel !l;
  output_value channel (arcs a.g);
  output_value channel a.initial;
  output_value channel a.final
;;

let input_automate channel =
  let g = ref(new_graphe())   in
  let l = input_value channel in
  do_list (fun (x,y) -> g := add_noeud !g x y) l;
  let l = input_value channel in
  do_list (fun (x,y) -> g := add_arc !g x y) l;
  let i = input_value channel in
  let f = input_value channel in
  {g = !g; initial = i; final = f}
;;

                    (* +-----------------------+
                       |  Affichage graphique  |
                       +-----------------------+ *)

let string_of_lettres(liste) =
  let lg = 2*list_length(liste) - 1 in
  let s = make_string lg `|`      in
  let i = ref 0                   in
  do_list (function Epsilon   -> s.[!i] <- `@`; i := !i+2
                  | Lettre(l) -> s.[!i] <- l;   i := !i+2) liste;
  s
;;

let green = rgb 0 192 0
and gris  = rgb 208 208 208
;;

let xgraphe_of_automate(a) =

  make_xgraphe
   (fun (e,s) -> (if list__mem e a.initial then green else blue),
                 (if list__mem e a.final   then -20   else 20),
                 s)
   (fun (_,l) -> (red, 10, string_of_lettres(l)))
   100
   a.g
   (hd (if a.initial <> [] then a.initial else noeuds(a.g)))
;;


                        (* +--------------+
                           |  Simulation  |
                           +--------------+ *)



let rec simule a xg cl tats =

  let cr = current_point()
  and encore = ref true
  and xn = map (etiq xg) tats in

  while !encore do

    (* colore les tats courants et attend une touche *)
    do_list (draw_xnoeud gris) xn;
    let ev = wait_next_event [Key_pressed] in
    do_list (draw_xnoeud background) xn;
    
    match ev.key with
    | `\027` -> raise Exit       (* Esc -> termin *)
    | `\008` -> encore := false  (* BS -> retourne en arrire *)
    | k -> moveto cr;            (* autre touche : avance *)
  	   set_color foreground;
  	   draw_char k;
  	   simule a xg cl (avance a cl tats (Lettre k));
  	   (* au retour, efface la touche et recommence *)
  	   set_color background;
  	   let (dx,dy) = text_size (make_string 1 k) in
  	   fill_rect cr dx dy
  done
;;
  

let simulation a cl init xg =

  clrscr "simulation: ";
  let cr = current_point() in
  display xg;
  begin
    try while true do moveto cr; simule a xg cl init done
    with Exit -> ()
  end;
  xg
  
;;


let show(a) =
   let cl = cltures(a) in
   edit 25 (xgraphe_of_automate a)
     "n(noeuds) l(lgendes) c(contrle) p(postscript) s(simulation) <Esc>"
     [`s`,simulation a cl (avance a cl a.initial Epsilon)]
;;

let xshow a =
  let ch = unix__open_process_out "showauto" in
  output_automate ch a;
  close_out ch
;;
