(* +------------------------------------------------------------------------+
   |                                                                        |
   |                                                                        |
   |                Manipulation et visualisation de graphes                |
   |                                                                        |
   |                                                                        |
   +------------------------------------------------------------------------+ *)
(* M. Quercia, le 13/04/98 *)

                        (* +------------------------+
                           |  Tableaux associatifs  |
                           +------------------------+ *)

type ('a,'b) table = ('a,'b) Mappe.t;;

let get    t a     = Mappe.find a t
and oget   t a     = try Some(Mappe.find a t) with Not_found -> None
and add    t (a,b) = Mappe.add a b (Mappe.remove a t)
and remove t a     = Mappe.remove a t
and set    t (a,b) = t := Mappe.add a b (Mappe.remove a !t)
;;

let ( <+> ) t (a,b) = add    t (a,b)
and ( <-> ) t  a    = remove t a
and ( <-- ) t (a,b) = set    t (a,b)
;;

let list_of_table(t) =
  let l = ref [] in
  Mappe.iter (fun x y -> l := (x,y) :: !l) t;
  !l
and table_of_list l =
  let t = ref(Mappe.empty compare) in
  List.iter (fun x -> t <-- x) l;
  !t
;;

let do_table f t = Mappe.iter (fun x y -> f(x,y)) t
and map_table f t =
  let t' = ref(table_of_list []) in
  Mappe.iter (fun x y -> t' <-- (f(x,y))) t;
  !t'
;;

                         (* +-------------+
                            |  Ensembles  |
                            +-------------+ *)

type 'a set = ('a,unit) Mappe.t

let set_of_list l =
  let s = ref (Mappe.empty compare) in
  List.iter (fun x -> s := Mappe.add x () !s) l;
  !s
;;

let list_of_set(s) =
  let l = ref [] in
  Mappe.iter (fun x y -> l := x :: !l) s;
  !l
;;

let do_set f = Mappe.iter (fun x y -> f(x))
and add s x  = Mappe.add x () s
and del s x  = Mappe.remove x s
;;


                     (* +---------------------+
                        |  Graphes abstraits  |
                        +---------------------+ *)

type ('a,'b,'c) graphe = {
  etiq : ('a,'b) table;
  succ : ('a,'a set) table;
  pred : ('a,'a set) table;
  arc  : ('a*'a,'c) table
};;

let new_graphe() = {
  etiq = table_of_list [];
  succ = table_of_list [];
  pred = table_of_list [];
  arc  = table_of_list []
};;

let add_noeud g n e = {
  etiq = g.etiq <+> (n,e);
  succ = g.succ <+> (n,match oget g.succ n with None -> set_of_list [] | Some(s) -> s);
  pred = g.pred <+> (n,match oget g.pred n with None -> set_of_list [] | Some(s) -> s);
  arc  = g.arc
};;

let del_noeud g n =
  let s = ref(g.succ) and p = ref(g.pred) and a = ref(g.arc) in
  do_set (fun x -> s <-- (x, del (get !s x) n); a := !a <-> (x,n)) (get g.pred n);
  do_set (fun x -> p <-- (x, del (get !p x) n); a := !a <-> (n,x)) (get g.succ n);
  {etiq = g.etiq <-> n; succ = !s <-> n; pred = !p <-> n; arc  = !a}
;;

let add_arc g (n,p) e = {
  etiq = g.etiq;
  succ = g.succ <+> (n, add (get g.succ n) p);
  pred = g.pred <+> (p, add (get g.pred p) n);
  arc  = g.arc  <+> ((n,p),e)
};;

let del_arc g (n,p) = {
  etiq = g.etiq;
  succ = g.succ <+> (n, del (get g.succ n) p);
  pred = g.pred <+> (p, del (get g.pred n) p);
  arc  = g.arc <-> (n,p)
};;

let noeuds(g) = List.map fst (list_of_table g.etiq)
and arcs(g)   = list_of_table(g.arc)
and member g x = try get g.etiq x; true with Not_found -> false
and succ  g x = list_of_set(get g.succ x)
and pred  g x = list_of_set(get g.pred x)
and etiq  g x = get g.etiq x
and arc   g (n,p) = get g.arc (n,p)
and arc_opt g (n,p) = try Some(get g.arc (n,p)) with Not_found -> None
;;

let do_noeuds f g = do_table f g.etiq
and do_arcs   f g = do_table f g.arc
;;


let mirroir(g) = { succ = g.pred;
                   pred = g.succ;
                   etiq = g.etiq;
                   arc  = map_table (fun ((x,y),c) -> ((y,x),c)) g.arc };;

                         (* +-------------+
                            |  Affichage  |
                            +-------------+ *)

open Graphx;;

type xnoeud = {
 npos   : point;     (* position d'affichage *)
 ncoul  : color;     (* couleur d'affichage  *)
 radius : int;       (* rayon du cercle (pas de cercle si =0, cercle double si <0) *)
 nlpos  : point;     (* position de la lgende *)
 nltxt  : string     (* texte de la lgende, pas de lgende si "" *)
}
and xarc = {
 apos  : point array; (* points de Bzier dfinissant le trac *)
 acoul : color;       (* couleur d'affichage *)
 arrow : int;         (* taille pointe de la flche (pas de pointe si <=0) *)
 alpos : point;       (* position de la lgende *)
 altxt : string       (* texte de la lgende, pas de lgende si "" *)
};;

type 'a xgraphe = ('a,xnoeud,xarc) graphe;;

let gris = rgb 192 192 192;;
let pi180 = 4.0*.atan(1.0)/.180.0;;

(* vecteur de norme r faisant un angle theta avec pq *)
let dir p q r theta =
  let dx = float(q.x - p.x)
  and dy = float(q.y - p.y)      in
  let f  = r /. sqrt(dx*.dx +. dy*.dy)  in
  let t  = float(theta) *. pi180 in
  let c  = f*.cos(t) and s = f*.sin(t)  in
  {x = truncate(c*.dx -. s*.dy); y = truncate(s*.dx +. c*.dy)}
;;
  
(* distance euclidienne *)
let lg p q =
  let dx = float(q.x - p.x)
  and dy = float(q.y - p.y) in
  sqrt(dx*.dx +. dy*.dy)
;;

(* translation *)
let ( ++ ) p q = {x = p.x + q.x; y = p.y + q.y}
and ( -- ) p q = {x = p.x - q.x; y = p.y - q.y};;

(* trace une lgende centre *)
let draw_legende p texte = if String.length(texte) > 0 then begin
  let (dx,dy) = text_size(texte) in
  moveto {x=p.x - dx/2; y=p.y - dy/2};
  draw_string texte
end;;

(* trace les points de contrle d'un arc *)
let draw_ctl(a) =
  set_color gris;
  draw_poly a.apos;
  set_color a.acoul;
  for i = 1 to Array.length(a.apos)-2 do fill_circle a.apos.(i) 5 done
;;

(* trace un arc et sa lgende *)
let draw_xarc r r' a =

  set_color a.acoul;

  (* courbe de Bzier : racourcit les segments de r et r' *)
  if a.arrow >= 0 then begin
  
    let poly = poly_of_bezier 4 a.apos  in
    let d  = ref(float(abs r))
    and d' = ref(float(abs r'))
    and i  = ref 0
    and j  = ref(Array.length(poly) - 1) in
  
    (* origine *)
    while (!i < !j) & (!d > 0.0) do
      let l = lg poly.(!i) poly.(!i+1) in
      if l < !d
      then begin d := !d -. l; incr i end
      else begin
	let v = dir poly.(!i+1) poly.(!i) (l -. !d) 0 in
	poly.(!i) <- poly.(!i+1) ++ v;
	d := 0.0
      end
    done;
  
    (* extrmit *)
    while (!i < !j) & (!d' > 0.0) do
      let l = lg poly.(!j) poly.(!j-1) in
      if l < !d'
      then begin d' := !d' -. l; decr j end
      else begin
	let v = dir poly.(!j-1) poly.(!j) (l -. !d') 0 in
	poly.(!j) <- poly.(!j-1) ++ v;
	d' := 0.0
      end
    done;
  
    (* trace l'arc *)
    if !i < !j then draw_poly (Array.sub poly !i (!j - !i + 1));
  
    (* pointe *)
    if !i < !j & a.arrow > 0 then begin
      let p = poly.(Array.length(poly)-1)
      and q = poly.(!j-1)
      and d = float a.arrow in
      let u = dir q p d 150
      and v = dir q p d (-150) in
      fill_poly [| poly.(!j); poly.(!j) ++ u; poly.(!j) ++ v |]
    end
  end;

  (* lgende *)
  draw_legende a.alpos a.altxt
;;

(* trace un noeud et sa lgende *)
let draw_xnoeud fond n =

  (* disque *)
  if n.radius <> 0 then begin
    let r = abs(n.radius) in
    set_color fond;
    fill_circle n.npos r;
    set_color n.ncoul;
    draw_circle n.npos r;
    if n.radius < 0 then draw_circle n.npos (r-3)
  end
  else set_color n.ncoul;
  
  (* lgende *)
  draw_legende n.nlpos n.nltxt
;;

(* trace les arcs et les noeuds *)
let xdisplay(xnoeuds,xarcs) =
  do_table (fun ((x,y),a) -> draw_xarc (get xnoeuds x).radius (get xnoeuds y).radius a) xarcs;
  do_table (fun (_,n) -> draw_xnoeud background n) xnoeuds
;;
let display(g) = xdisplay(g.etiq,g.arc);;

(* place les noeuds par ordre topologique *)
let place_noeuds f pas gr n =

  let pos    = ref(table_of_list [])          (* positions          *)
  and haut   = ref(table_of_list [])          (* hauteur par niveau *)
  and hmax   = ref(0)                         (* hauteur maxi       *)
  and (x,y)  = (ref 0, ref 0)                 (* dernire position  *)
  and file   = Queue.create() in              (* file d'attente     *)

  (* place un noeud et sa composante connexe *)
  let place(n) = if oget !pos n = None then begin
    pos <-- (n,(!x,!y));
    Queue.add n file;
  
    (* parcours du graphe en largeur *)
    begin try while true do
  
	let n = Queue.take file     in     (* noeud suivant *)
	let p = fst(get !pos n) + 1 in     (* abscisse fils *)
  
	if  p > !x then begin              (* chgt niveau ?   *)
	  haut <-- (!x, !y);               (* note la hauteur *)
	  hmax := max !hmax !y;
	  incr x;
	  y := 0
	end;
  
	let job(a) = if oget !pos a = None then begin (* placement d'un fils *)
	  pos <-- (a,(!x,!y));
	  incr y;
	  Queue.add a file
	end
	  
	in
	(match oget gr.succ n with Some(l) -> do_set job l | None -> ());
	(match oget gr.pred n with Some(l) -> do_set job l | None -> ())
      done
      with Queue.Empty -> haut <-- (!x,!y);     (* dernire hauteur *)
			   hmax := max !hmax !y
  
    end
  end
  in List.iter place (n :: noeuds(gr));

  (* centre les niveaux verticalement et constitue les champs xnoeuds *)
  let job(n,(x,y)) =
    let p = {x = (x + 1)*pas; y = (y + (!hmax - (get !haut x))/2 + 1)*pas} in
    let (couleur,radius,txt) = f(n, get gr.etiq n) in
    (n, { npos   = p;
          ncoul  = couleur;
          radius = radius;
          nlpos  = p;
          nltxt  = txt})
  in map_table job !pos
;;

(* dcale la position d'une lgende *)
(* p = point initial                *)
(* d = direction tangente           *)
(* txt = texte de la lgende        *)
let move d p txt =
  let (dx,dy) = text_size(txt) in
  match 0 with
  | _ when d.x = 0 & d.y > 0 -> {x=p.x+dx/2+3; y=p.y}
  | _ when d.x = 0           -> {x=p.x-dx/2-3; y=p.y}
  | _ when d.y = 0 & d.x > 0 -> {x=p.x; y=p.y+dy/2+2}
  | _ when d.y = 0           -> {x=p.x; y=p.y-dy/2-2}
  | _ when d.x > 0 & d.y > 0 -> {x=p.x-dx/2-3; y=p.y+3}
  | _ when d.x > 0           -> {x=p.x+dx/2+3; y=p.y+3}
  | _ when           d.y > 0 -> {x=p.x-dx/2-3; y=p.y-3}
  | _                        -> {x=p.x+dx/2+3; y=p.y-3}
;;


(* cherche les noeuds  proximit d'un segment et retourne la distance  viter *)
let mindist p q xnoeuds =
  let d = ref 0
  and l = lg p q in
  do_table (fun (_,x) -> if p <> x.npos & q <> x.npos
                         & (lg p x.npos) +. (lg x.npos q) -. l < 1.0
                         then d := max !d (abs x.radius))
           xnoeuds;
  2.0*.(float !d)
;;

(* calcule les arcs  partir des noeuds *)
let place_arcs g xnoeuds arcs =

  (* calcule les points de Bzier d'un arc *)
  let arc((x,y),e) =
    let p = (get xnoeuds x).npos and q = (get xnoeuds y).npos
    and (couleur,arrow,txt) = g((x,y),e) in

    if p = q then begin (* trace une boucle verticale *)
      let d = (let d = (get xnoeuds x).radius in if d > 0 then d else 20) in
      ((x,y), (* 26/5 = 6sqrt(3) *)
      { apos  = [| p; p ++ {x=(-3*d); y=(26*d)/5}; p ++ {x=(3*d); y=(26*d)/5}; p |];
        acoul = couleur;
        arrow = arrow;
        alpos = move {x=1; y=0} (p ++ {x=0; y=4*d}) txt;
        altxt = txt })
    end
    else match oget arcs (y,x) with

    | None -> (* pas d'arc retour, trace un segment s'il n'y a pas de *)
              (* noeud au milieu, sinon un arc.                       *)
      let d = mindist p q xnoeuds                                  in
      let r = {x = (p.x+q.x)/2; y = (p.y+q.y)/2} ++ (dir q p d 90) in
      ((x,y),
      { apos  = [| p; r; r; q |];
        acoul = couleur;
        arrow = arrow;
        alpos = move (q -- p) r txt;
        altxt = txt })

    | Some(_) -> (* arc retour, trace une courbe *)
      let d = (let d = max (get xnoeuds x).radius (get xnoeuds y).radius in
               if d > 0 then d else 20) in
      let r = {x = (p.x+q.x)/2; y = (p.y+q.y)/2} ++ (dir p q (float d) 90) in
      ((x,y),
      { apos  = [| p; r; r; q |];
        acoul = couleur;
        arrow = arrow;
        alpos = move (q -- p) r txt;
        altxt = txt })

  in
  map_table arc arcs
;;

let make_xgraphe f g pas gr n =

  let xnoeuds = place_noeuds f pas gr n in
  let xarcs   = place_arcs   g xnoeuds gr.arc in
  { etiq = xnoeuds;
    succ = gr.succ;
    pred = gr.pred;
    arc  = xarcs }
;;

                    (* +----------------------+
                       |  Edition des noeuds  |
                       +----------------------+ *)

(* efface l'cran et affiche un texte en {x=0; y=0} *)
let clrscr(txt) =
  clear_graph();
  set_color foreground;
  moveto {x=0; y=0};
  draw_string txt
;;

(* trace une grille *)
let draw_grille(pas) = if pas > 0 then begin
   set_color gris;
   for i=1 to (size_x()-1)/pas do draw_poly [|{x=i*pas;y=0}; {x=i*pas; y=size_y()}|] done;
   for i=1 to (size_y()-1)/pas do draw_poly [|{x=0;y=i*pas}; {x=size_x(); y=i*pas}|] done
end;;

(* slectionne un noeud *)
let select_noeud(g) =

  let noeud = ref(None)
  and ev    = wait_next_event [Button_down; Key_pressed]
  in if ev.button
  then do_table
    (fun (n,x) -> if lg ev.mouse x.npos <= float(abs x.radius)
                  then noeud := Some(n))
     g.etiq;
  (ev,!noeud)
;;

(* dplace un noeud *)
let move_noeud pas n ({etiq=xnoeuds; arc=xarcs} as g) =

  (* arrondi *)
  let arrondi p =
    if pas <= 0 then p
    else {x = pas*((p.x+pas/2)/pas); y = pas*((p.y+pas/2)/pas)}
  in

  (* mesure les dcalages entre le noeud, la souris et la lgende *)
  let xn = get xnoeuds n in
  let d  = xn.npos -- mouse_pos()
  and dl = xn.nlpos -- xn.npos in

  (* spare le graphe en deux *)
  let ynoeuds = ref(xnoeuds <-> n)
  and zn      = ref(xn)
  and yarcs   = ref(table_of_list [])
  and zarcs   = ref(table_of_list []) in
  do_table (fun ((a,b),_ as arc) -> (if n=a or n=b then zarcs else yarcs) <-- arc) xarcs;

  (* enregistre le fond *)
  clrscr "move_noeuds ...";
  draw_grille(pas);
  xdisplay(!ynoeuds, !yarcs);
  let fond = get_image 0 0 (size_x()) (size_y()) in

  (* mise  jour affichage *)
  let maj(p) =
    zn := {npos   = p;
           ncoul  = foreground;
           radius = xn.radius;
           nlpos  = p ++ dl;
           nltxt  = xn.nltxt};
    zarcs := place_arcs (fun (a,ya) -> (ya.acoul,ya.arrow,ya.altxt))
                        (xnoeuds <+> (n,!zn)) !zarcs;
    draw_image fond 0 0;
    do_table 
      (fun ((x,y),a) -> draw_xarc (get xnoeuds x).radius (get xnoeuds y).radius a)
      !zarcs;
    draw_xnoeud background !zn
  in
  
  (* boucle de mouvement *)
  maj(!zn.npos);
  let encore = ref true in
  while !encore do
    match wait_next_event [Button_up; Mouse_motion] with
    | {button=false}         -> encore := false
    | {button=true; mouse=m} ->
        let p = arrondi(m ++ d) in if p <> !zn.npos then maj(p)
  done;

  (* position finale *)
  ynoeuds <-- (n,{npos   = !zn.npos;
                  ncoul  = xn.ncoul;
                  radius = xn.radius;
                  nlpos  = !zn.nlpos;
                  nltxt  = xn.nltxt});
  do_table (fun ((a,b),c) -> let arc = get xarcs (a,b) in
            yarcs <-- ((a,b),{apos  = c.apos;
                              acoul = arc.acoul;
                              arrow = arc.arrow;
                              alpos = c.alpos;
                              altxt = arc.altxt})) !zarcs;
  {etiq = !ynoeuds; arc = !yarcs; succ = g.succ; pred = g.pred}
;;

(* dplace plusieurs noeuds *)
let move_noeuds pas g =

  let g = ref g in
  clrscr "move_noeuds ...";
  draw_grille(pas);
  display(!g);

  let ev = ref(select_noeud !g) in
  while not((fst !ev).keypressed) do
    begin match snd !ev with
    | Some(n) -> g := move_noeud pas n !g;
                 clrscr "move_noeuds ...";
   		 draw_grille(pas);
                 display(!g)
    | None -> ()
    end;
    ev := select_noeud(!g)
  done;

  !g
;;

                   (* +------------------------+
                      |  Edition des lgendes  |
                      +------------------------+ *)

(* slectionne une lgende *)
let select_leg(g) =

  let noeud = ref(None)
  and arc   = ref(None)
  and ev    = wait_next_event [Button_down; Key_pressed]
  in if ev.button then begin
    let m = ev.mouse in
    do_table (fun (n, {nlpos=p; nltxt=t}) ->
                  let (dx,dy) = text_size(t) in
                  if t <> "" & abs(p.x-m.x) <= (dx+1)/2 & abs(p.y-m.y) <= (dy+1)/2
                  then noeud := Some(n))
             g.etiq;
    do_table (fun (a, {alpos=p; altxt=t}) ->
                  let (dx,dy) = text_size(t) in
                  if t <> "" & abs(p.x-m.x) <= (dx+1)/2 & abs(p.y-m.y) <= (dy+1)/2
                  then arc   := Some(a))
             g.arc
  end;
  (ev,(!noeud,!arc))
;;

(* dplace une lgende  la souris *)
let move_leg pos txt =

  let p = ref(pos)
  and d = pos -- mouse_pos() in
  let fond = get_image 0 0 (size_x()) (size_y()) in
  set_color foreground;
  draw_legende !p txt;

  let encore = ref true in
  while !encore do
    match wait_next_event [Button_up; Mouse_motion] with
    | {button=false}         -> encore := false
    | {button=true; mouse=m} ->
        let q = m ++ d in
        if q <> !p then begin
          p := q;
          draw_image fond 0 0;
          draw_legende !p txt
        end
  done;
  !p
;;

(* dplace plusieurs lgendes *)
let move_legs(g) =

  let g = ref(g) in
  clrscr "move_legs ...";
  display(!g);

  let ev = ref(select_leg !g) in
  while not((fst !ev).keypressed) do
    begin match snd !ev with

    | (Some n,_) ->
      let xn = get (!g.etiq) n in
      clrscr "move_legs ...";
      xdisplay(!g.etiq <+> (n,{npos   = xn.npos;
	      	 	        ncoul  = xn.ncoul;
	      		     	radius = xn.radius;
	      		    	nlpos  = xn.nlpos;
	      		    	nltxt  = ""}),
	       !g.arc);
      g := {arc = !g.arc; succ = !g.succ; pred = !g.pred;
            etiq = !g.etiq <+> (n,{npos   = xn.npos;
		      		   ncoul  = xn.ncoul;
		      		   radius = xn.radius;
		      		   nlpos  = move_leg xn.nlpos xn.nltxt;
		      		   nltxt  = xn.nltxt})};
      display(!g)

    | (_,Some a) ->
      let xa = get (!g.arc) a in
      clrscr "move_legs ...";
      xdisplay(!g.etiq,
	       !g.arc <+> (a,{apos  = xa.apos;
			      acoul = xa.acoul;
			      arrow = xa.arrow;
			      alpos = xa.alpos;
			      altxt = ""}));
      g := {etiq = !g.etiq; succ = !g.succ; pred = !g.pred;
             arc = !g.arc <+> (a,{apos  = xa.apos;
		    		  acoul = xa.acoul;
		    		  arrow = xa.arrow;
		    		  alpos = move_leg xa.alpos xa.altxt;
		    		  altxt = xa.altxt})};
      display(!g)

    | _ -> ()
    end;
    ev := select_leg(!g)
  done;

  !g
;;


              (* +----------------------------------+
                 |  Edition des points de contrle  |
                 +----------------------------------+ *)

(* slectionne un point de contrle *)
let select_ctl(g) =

  let point = ref(None)
  and ev    = wait_next_event [Button_down; Key_pressed]
  in if ev.button then begin
    let m = ev.mouse in
    do_table (fun (a, {apos=p}) ->
                  for i=1 to Array.length(p)-2 do
                  if lg m p.(i) < 10.0 then point := Some(a,i) done)
             g.arc
  end;
  (ev,!point)
;;

(* divise un arc en deux *)
let add_ctl(a,i) =
  let j = i - (i mod 3) in
  let dbut  = Array.sub !a.apos 0 j
  and fin    = Array.sub !a.apos (j+4) (Array.length(!a.apos) - j - 4)
  and milieu = poly_of_bezier 1 (Array.sub !a.apos j 4)
  in
  a := {apos  = Array.append (Array.append dbut milieu) fin;
        acoul = !a.acoul;
        arrow = !a.arrow;
        alpos = !a.alpos;
        altxt = !a.altxt};
  false
;;

(* supprime un point et ses deux voisins *)
let del_ctl(a,i) =

  if i-1 <= 0 or i+1 >= Array.length(!a.apos) then true
  else begin
    let dbut  = Array.sub !a.apos 0 (i-1)
    and fin    = Array.sub !a.apos (i+2) (Array.length(!a.apos) - i - 2) in
    a := {apos  = Array.append dbut fin;
          acoul = !a.acoul;
          arrow = !a.arrow;
          alpos = !a.alpos;
          altxt = !a.altxt};
    false
  end
;;

(* dplace un point de contrle *)
let move_ctl a i (xnoeuds,xarcs) =

  (* mesure le dcalage entre la souris et le point *)
  let xa = get xarcs a in
  let d  = xa.apos.(i) -- mouse_pos() in

  (* isole l'arc  modifier *)
  let yarcs = xarcs <-> a
  and ya    = ref {apos  = Array.copy xa.apos;
                   acoul = foreground;
                   arrow = xa.arrow;
                   alpos = xa.alpos;
                   altxt = xa.altxt}
  and (k,l)  = let lg = Array.length xa.apos in (lg/2-1, lg/2+(lg mod 2))
  and (r,r') = ((get xnoeuds (fst a)).radius,(get xnoeuds (snd a)).radius)
  in

  (* enregistre le fond *)
  clrscr "move_ctls ...";
  xdisplay(xnoeuds,yarcs);
  do_table (fun (_,a) -> draw_ctl a) yarcs;
  let fond = get_image 0 0 (size_x()) (size_y()) in

  (* mise  jour affichage *)
  let maj(p) =
    !ya.apos.(i) <- p;
    let u = !ya.apos.(k) and v = !ya.apos.(l) in
    let q = move (v--u) {x=(v.x+u.x)/2; y=(v.y+u.y)/2} !ya.altxt in
    ya := { apos  = !ya.apos;
            acoul = foreground;
            arrow = !ya.arrow;
            alpos = q;
            altxt = !ya.altxt };
    draw_image fond 0 0;
    draw_xarc  r r' !ya;
    draw_ctl   !ya
  in

  (* boucle de mouvement *)
  maj(!ya.apos.(i));
  let encore = ref true in
  while !encore do
    match wait_next_event [Button_up; Mouse_motion; Key_pressed] with
    | {button=false}           -> encore := false
    | {keypressed=true; key=k} -> begin match k with
        | ('+' | 'a') -> encore := add_ctl(ya,i)
        | ('-' | 's') -> encore := del_ctl(ya,i)
        | _ -> ()
      end
    | {mouse=m} -> let p = m ++ d in if p <> !ya.apos.(i) then maj(p)
  done;

  (* position finale *)
  ya := {apos   = !ya.apos;
         acoul  = xa.acoul;
         arrow  = !ya.arrow;
         alpos  = !ya.alpos;
         altxt  = !ya.altxt};
  draw_image fond 0 0;
  draw_ctl   !ya;
  draw_xarc  r r' !ya;
  yarcs <+> (a,!ya)
  
;;

(* dplace plusieurs points de contrle *)
let move_ctls(g) =

  let g = ref(g) in
  clrscr "move_ctls ...";
  display(!g);
  do_table (fun (_,a) -> draw_ctl a) !g.arc;

  let ev = ref(select_ctl(!g)) in
  while not((fst !ev).keypressed) do
    begin match snd !ev with
    | Some(a,i) -> g := {etiq = !g.etiq; succ = !g.succ; pred = !g.pred;
                         arc = move_ctl a i (!g.etiq,!g.arc)};
    | _ -> ()
    end;
    ev := select_ctl(!g)
  done;

  !g
;;

                  (* +-------------------------------------+
                     |  Saisie d'une chaine de caractres  |
                     +-------------------------------------+ *)

let get_string(prompt,init) =

  (* efface la zone d'dition *)
  let (_,dy) = text_size "A" in
  set_color background;
  fill_rect {x=0; y=0} (size_x()) (dy+10);
  set_color foreground;

  (* affiche le prompt *)
  moveto {x=0; y=5};
  draw_string prompt;
  let p = current_point() in

  (* boucle d'dition *)
  let dbut  = ref init
  and fin    = ref ""
  and encore = ref true in
  while !encore do

    (* raffiche la chaine et le curseur *)
    moveto p;
    set_color background;
    fill_rect {x=p.x-5; y=p.y-5} (size_x()) (dy+10);
    set_color foreground;
    draw_string !dbut;
    let q = current_point() in
    fill_poly [|{x=q.x;   y=q.y+2};
                {x=q.x-5; y=q.y-5};
                {x=q.x+5; y=q.y-5};
                {x=q.x;   y=q.y+2}|];
    draw_string !fin;

    match (wait_next_event[Key_pressed]).key with
    | (* esc *) '\027' -> raise Exit
    | (* ret *) '\013' -> encore := false
    | (* BS  *) '\008' ->
      if !dbut <> ""
      then dbut := String.sub !dbut 0 (String.length !dbut - 1)

    | (* del *) ('\004'|'\127') ->
      if !fin <> ""
      then fin := String.sub !fin 1 (String.length !fin - 1)

    | (* ^B *) '\002' ->
      if !dbut <> "" then begin
         fin   := (String.make 1 !dbut.[String.length !dbut - 1]) ^ !fin;
         dbut := String.sub !dbut 0 (String.length !dbut - 1)
      end

    | (* ^F *) '\006' ->
      if !fin <> "" then begin
         dbut := !dbut ^ (String.make 1 !fin.[0]);
         fin   := String.sub !fin 1 (String.length !fin - 1)
      end

    | (* ^U *) '\021' -> dbut := ""
    | (* ^K *) '\011' -> fin   := ""
    | (* ^A *) '\001' -> fin := !dbut ^ !fin; dbut := ""
    | (* ^E *) '\005' -> dbut := !dbut ^ !fin; fin := ""

    | c -> dbut := !dbut ^ (String.make 1 c)

  done;

  !dbut ^ !fin
;;

                 (* +----------------------------+
                    |  Conversion en postscript  |
                    +----------------------------+ *)

let postscript(g) = try
  let nom = get_string("fichier postscript: ","essai.ps") in
  clear_graph(); display(g);
  write_postscript(nom)
with Exit -> ()
;;


                     (* +---------------------+
                        |  Editeur gnrique  |
                        +---------------------+ *)

let edit pas xg msg usermenu =

  let xg = ref(xg)
  and msg = (if msg = "" then "Menu: n(noeuds) l(lgendes) c(contrle) p(postscript) <Esc>"
                         else msg)
  in

  clrscr msg;
  display(!xg);

  let ev = ref(wait_next_event [Key_pressed]) in
  while !ev.key <> '\027' do
    begin match !ev.key with
    | k when List.mem_assoc k usermenu -> xg := List.assoc k usermenu !xg
    | 'n' -> xg := move_noeuds pas !xg
    | 'l' -> xg := move_legs       !xg
    | 'c' -> xg := move_ctls       !xg
    | 'p' -> postscript !xg
    | _   -> ()
    end;
    clrscr msg;
    display(!xg);
    ev := wait_next_event [Key_pressed]
  done;
  clear_graph();
  display(!xg);
  !xg

;;
